# Copyright (c) 2006 Tilman Sauerbeck (tilman at code-monkey de)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of version 2 of the GNU General Public License as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

require "ecore"
require "ecore_con"

module Embrace; end

module Embrace::IMAP
	class IMAPError < StandardError; end
	class LoginError < IMAPError; end

	class MailboxStatusEvent < Ecore::Event
		attr_reader :name, :count

		def initialize(name, count)
			super()

			@name, @count = name, count
		end
	end

	class FinishedEvent < Ecore::Event; end

	class Session
		attr_reader :connection

		def initialize(config)
			@login = config[:server][:login]
			@password = config[:server][:password]

			flags = 2
			flags |= 16 if config[:server][:use_ssl]

			@connection = Ecore::Con::Server.new(flags,
			                                     config[:server][:host],
			                                     config[:server][:port])
			@buffer = ""

			@tag_id = 0
			@requests = []

			@mboxes = config[:mailboxes].dup
			@state = :disconnected

			@handlers = [
				Ecore::EventHandler.new(Ecore::Con::ServerAddEvent,
				                        &method(:on_add)),
				Ecore::EventHandler.new(Ecore::Con::ServerDataEvent,
				                        &method(:on_data)),
				Ecore::EventHandler.new(Ecore::Con::ServerDelEvent,
				                        &method(:on_del))
			]
		end

		def next_tag
			@tag_id = 0 if @tag_id == 0xffff

			"0x%x" % [@tag_id += 1]
		end

		private
		def login(login, password)
			@requests << LoginRequest.new(self, login, password).send
		end

		def logout
			@requests << LogoutRequest.new(self).send
		end

		def query_status(mailbox)
			@requests << StatusRequest.new(self, mailbox).send
		end

		def on_add(ev)
			return true unless ev.server == @connection

			@state = :connected

			false
		end

		def on_data(ev)
			return true unless ev.server == @connection

			lines = (@buffer + ev.data).split(/\r\n/, -1)
			@buffer = lines.pop

			lines.each { |line| handle_line(line.strip) }

			false
		end

		def on_del(ev)
			return true unless ev.server == @connection

			@handlers.each { |h| h.delete }

			@state = :disconnected
			@connection.delete

			FinishedEvent.raise

			false
		end

		def handle_line(line)
if $DEBUG
			puts("<- " + line)
end
			handle_response(Response.deserialize(line))

			if @state == :connected
				login(@login, @password)
			end
		end

		def handle_response(resp)
			unless resp.tagged?
				handle_untagged_response(resp)
			else
				req = @requests.find { |r| r.tag == resp.tag }
				handle_tagged_response(resp, req)
				@requests.delete(req)
			end
		end

		def handle_tagged_response(resp, req)
			return unless req.is_a?(LoginRequest)

			case resp.status
			when :ok
				@state = :logged_in

				@mboxes.each { |mb| query_status(mb) }
			else
				raise(LoginError, "cannot login - #{resp.data}")
			end
		end

		def handle_untagged_response(resp)
			return if resp.key != "STATUS"

			md = resp.data.match(/\"(.+)\" \(UNSEEN (\d+)\)/)
			if md.nil?
				raise(IMAPError, "invalid status response - #{resp.data}")
			end

			name, count = md.captures.first, md.captures.last.to_i

			MailboxStatusEvent.raise(name, count)
			@mboxes.delete(name)

			if @mboxes.empty?
				@status = :done
				logout
			end
		end
	end

	class Request
		attr_reader :tag

		def initialize(session)
			@session = session
			@tag = session.next_tag
		end

		def send
if $DEBUG
			puts("-> #{@tag} #{serialize}")
end
			@session.connection << "#{@tag} #{serialize}\r\n"

			self
		end
	end

	class LoginRequest < Request
		def initialize(session, login, password)
			super(session)

			@login = login
			@password = password
		end

		def serialize
			"LOGIN #{@login} #{@password}"
		end
	end

	class LogoutRequest < Request
		def serialize
			"LOGOUT"
		end
	end

	class StatusRequest < Request
		def initialize(session, mailbox)
			super(session)

			@mailbox = mailbox
		end

		def serialize
			"STATUS #{@mailbox} (UNSEEN)"
		end
	end

	class Response
		class ResponseError < IMAPError; end

		attr_reader :data

		def Response.deserialize(line)
			case line
			when /^\* (\w+) (.*)$/
				UntaggedResponse.new($1, $2)
			when /^(0x[[:xdigit:]]+) (OK|NO|BAD) /
				tag = $1
				rest = line[(tag.length + 1)..-1]
				status = $2.downcase.to_sym

				TaggedResponse.new(tag, status, rest)
			else
				raise(ResponseError, "cannot parse - #{line}")
			end
		end

		def initialize(data)
			@data = data.to_str.freeze
		end

		def tagged?
			false
		end
	end

	class TaggedResponse < Response
		attr_reader :tag, :status

		def initialize(tag, status, data)
			super(data)

			@tag = tag
			@status = status
		end

		def tagged?
			true
		end
	end

	class UntaggedResponse < Response
		attr_reader :key

		def initialize(key, data)
			super(data)

			@key = key
		end
	end
end
